import typing as tp

import pandas as pd

DF = pd.DataFrame


def validate_parameters(params: tp.List, params_names: tp.List[str], target_type: tp.Any) -> None:
    """Validates a list of parameters' types and values.

    Parameters
    ----------
    params
        Parameters to validate.
    params_names: str
        Name of the parameters to be printed in error messages.
    target_type
        Acceptable data types for the parameters.

    Raises
    ------
    TypeError
        If the parameter's type does not match the desired type.
    ValueError
        If the parameter is empty.
    """
    for param, param_name in zip(params, params_names):
        validate_parameter(param=param, param_name=param_name, target_type=target_type)


def validate_parameter(param: tp.Any, param_name: str, target_type: tp.Any) -> None:
    """Validates a parameter's type and value.

    Parameters
    ----------
    param
        Parameter to validate.
    param_name: str
        Name of the parameter to be printed in error messages.
    target_type
        Acceptable data types for the parameter.

    Raises
    ------
    TypeError
        If the parameter's type does not match the desired type.
    ValueError
        If the parameter is empty.
    """
    if not isinstance(param, target_type):
        raise TypeError(f'`{param_name}` must be an instance of {target_type}, not {type(param)}.')

    if not len(param):
        raise ValueError(f'`{param_name}` cannot be empty.')


def validate_dataframe(df: DF, columns: tp.List[str]) -> None:
    """Validates if `df` is not empty and has answer column.

    Parameters
    ----------
    df: pandas.DataFrame
        Dataframe.
    columns: list of str
        Required columns.

    Raises
    ------
    ValueError
        If the `df` is empty.
    AttributeError
        If any of the expected `columns` is not present in `df`.
    """
    if df.empty:
        raise ValueError('Dataframe cannot be empty.')

    for col in columns:
        validate_column(df=df, column=col)


def validate_column(df: DF, column: str) -> None:
    """Validates if `df` has `column`.

    Parameters
    ----------
    df: pandas.DataFrame
        Dataframe.
    column: str
        Target column.

    Raises
    ------
    AttributeError
        If any of the expected columns is not present in `df`.
    """
    if column not in df.columns:
        raise AttributeError(f'Dataframe does not contain required column `{column}`.')



__author__ = 'Milo Utsch'
__version__ = '0.1.0'

import pytest

import pandas as pd

from take_ai_evaluation.persistence import read_dataframe


def test_read_dataframe_from_file():
    pass


def test_read_dataframe_from_powerbi():
    mock_df = pd.DataFrame([{'Sentence': 'cadastra ai', 'Intent': 'cadastrar', 'Predicted': 'cadastrar'},
                            {'Sentence': 'compra pra mim', 'Intent': 'compra', 'Predicted': 'cadastrar'},
                            ])
    
    result = read_dataframe(mock_df)
    
    pd.testing.assert_frame_equal(result, mock_df, check_like=True, check_dtype=False)



import pandas as pd
from sklearn.preprocessing import normalize as norm

DF = pd.DataFrame


def normalize_confusion_matrix(confusion_matrix: DF) -> DF:
    """Calculates the normalized confusion matrix using L1 normalization.

    Parameters
    ----------
    confusion_matrix : pandas.DataFrame
        Dataframe containing the original confusion matrix.

    Returns
    -------
    matrix : pandas.DataFrame
        Dataframe containing the normalized confusion matrix.
    """
    return pd.DataFrame(data=norm(confusion_matrix, norm="l1"),
                        index=confusion_matrix.index,
                        columns=confusion_matrix.columns)



import pandas as pd

DF = pd.DataFrame


class Intent:
    """Wraps methods for calculating insights regarding the quality of the intent.

    Attributes
    ----------
    label : str
        Intent's label (name).
    tp: float
        Intent's true positive predictions.
    fn: float
        Intent's false negative predictions.
    fp: float
        Intent's false positive predictions.
    tn: float
        Intent's true negative predictions.
    one_vs_all_confusion_matrix: pandas.DataFrame
        ONE vs ALL confusion matrix of the intent.
    metrics_mapper : dict
        Dict that maps accuracy, precision, recall and f1 values for a intent.
    accuracy: float
        Accuracy score for a intent.
    precision : float
        Precision score for a intent.
    recall : float
        Recall score for a intent.
    f1 : float
        F1 score, also known as balanced F-score or F-measure, for a intent.

    Methods
    -------
    get_metric(metric)
        Provides the calculated value for a specific metric.

    Raises
    ------
    ZeroDivisionError
        If during metrics calculation the denominator evaluates to zero.
    """

    def __init__(self, label: str, all_vs_all_confusion_matrix: DF) -> None:
        """
        Parameters
        ----------
        label : str
            Label of the intent to be retrieved.
        all_vs_all_confusion_matrix : pandas.DataFrame
            ALL vs ALL confusion matrix.
        """
        self.label = label

        self.__make_one_vs_all_confusion_matrix(all_vs_all_confusion_matrix=all_vs_all_confusion_matrix)
        self.__make_metrics_mapper()

    def __create_one_vs_all_confusion_matrix(self) -> DF:
        """Creates a ONE vs ALL (multilabel) confusion matrix filled with it's results.

        Returns
        -------
        matrix : pandas.DataFrame
            Metrics filled 2d dataframe with the size of 2 by 2.
        """
        data = ((self.tp, self.fn), (self.fp, self.tn))

        return pd.DataFrame(data=data, columns=['Positive', 'Negative'], index=['Positive', 'Negative'])

    def __calculate_one_vs_all_confusion_matrix(self, all_vs_all_confusion_matrix: DF) -> None:
        """Calculates the confusion matrix (TP, FP, FN, TN) for a intent.

        Parameters
        ----------
        all_vs_all_confusion_matrix : pandas.DataFrame
            ALL vs ALL confusion matrix.
        """
        n = all_vs_all_confusion_matrix.sum().sum()
        self.tp = all_vs_all_confusion_matrix[self.label][self.label]
        self.fp = all_vs_all_confusion_matrix[self.label].sum() - self.tp
        self.fn = all_vs_all_confusion_matrix.loc[self.label].sum() - self.tp
        self.tn = n - (self.tp + self.fn + self.fp)

    def __make_one_vs_all_confusion_matrix(self, all_vs_all_confusion_matrix: DF) -> None:
        """Generates and assigns a confusion matrix for a intent.

        Parameters
        ----------
        all_vs_all_confusion_matrix : pandas.DataFrame
            ALL vs ALL confusion matrix.
        """
        self.__calculate_one_vs_all_confusion_matrix(all_vs_all_confusion_matrix=all_vs_all_confusion_matrix)
        self.one_vs_all_confusion_matrix = self.__create_one_vs_all_confusion_matrix()

    def __calculate_metrics(self) -> None:
        """Calculates accuracy, precision, recall and f1 for a intent.

        Raises
        ------
        ZeroDivisionError
            If during metrics calculation the denominator evaluates to zero.
        """
        self.__calculate_accuracy()
        self.__calculate_precision()
        self.__calculate_recall()
        self.__calculate_f1()

    def __make_metrics_mapper(self) -> None:
        """Calculates and maps accuracy, precision, recall and f1 for a intent.

        Raises
        ------
        ZeroDivisionError
            If during metrics calculation the denominator evaluates to zero.
        """
        self.__calculate_metrics()

        self.metrics_mapper = {'accuracy': self.accuracy,
                               'precision': self.precision,
                               'recall': self.recall,
                               'f1': self.f1}

    def __calculate_accuracy(self) -> None:
        """Calculates the accuracy score for a intent.

        The accuracy is the ratio `(tp + fn) / (tp + fn + tn + fp)`
        where `tp` is the number of true positives, `fn` the number
        of false negatives, `tn` is the number of true negatives
        and `fp` the number of false positives. The accuracy is the
        ability of the instrument to measure the accurate value.
        In other words, it is the closeness of the measured
        value to a standard or true value.

        The best value is 1 and the worst value is 0.

        Raises
        ------
        ZeroDivisionError
            If `(tp + fn + tn + fp)` evaluates to zero.
        """
        nominator = (self.tp + self.fn)
        denominator = (self.tp + self.fn + self.tn + self.fp)

        if denominator > 0:
            self.accuracy = nominator / denominator
        else:
            raise ZeroDivisionError(f"`{self.label}` accuracy denominator cannot be zero.")

    def __calculate_precision(self) -> None:
        """Calculates the precision score for a intent.

        The precision is the ratio `tp / (tp + fp)` where `tp` is
        the number of true positives and `fp` the number of false
        positives. The precision is intuitively the ability of the
        classifier not to label as positive a sample that is negative.

        The best value is 1 and the worst value is 0.

        Raises
        ------
        ZeroDivisionError
            If `(tp + fp)` evaluates to zero.
        """
        denominator = (self.tp + self.fp)

        if denominator > 0:
            self.precision = self.tp / denominator
        else:
            raise ZeroDivisionError(f"`{self.label}` precision denominator cannot be zero.")

    def __calculate_recall(self) -> None:
        """Calculate the recall score for a intent.

        The recall is the ratio `tp / (tp + fn)` where `tp` is
        the number of true positives and `fn` the number of false
        negatives. The recall is intuitively the ability of the
        classifier to find all the positive samples.

        The best value is 1 and the worst value is 0.

        Raises
        ------
        ZeroDivisionError
            If `(tp + fn)` evaluates to zero.
        """
        denominator = (self.tp + self.fn)

        if denominator > 0:
            self.recall = self.tp / denominator
        else:
            raise ZeroDivisionError(f"`{self.label}` recall denominator cannot be zero.")

    def __calculate_f1(self) -> None:
        """Calculates the F1 score, also known as balanced F-score or F-measure, for a intent.

        The F1 score is the ratio `2 * (precision * recall) / (precision + recall)`
        and can be interpreted as a weighted average of the precision
        and recall. The relative contribution of precision and recall
        to the F1 score are equal.

        The best value is 1 and the worst value is 0.

        Raises
        ------
        ZeroDivisionError
            If `(precision + recall)` evaluates to zero.
        """
        nominator = (2 * self.precision * self.recall)
        denominator = (self.precision + self.recall)

        if denominator > 0:
            self.f1 = nominator / denominator
        else:
            raise ZeroDivisionError(f"`{self.label}` F1 denominator cannot be zero.")

    def get_metric(self, metric: str) -> float:
        """Provides the calculated value for a specific metric of a intent.

        Parameters
        ----------
        metric : str
            Metric for which the value will be retrived.
            Can be either of "accuracy", "precision", "recall" or "f1".

        Returns
        -------
        metric : float
            Requested metric's value.

        Raises
        ------
        KeyError
            If `metric` is any of the expected ones.
        """
        return self.metrics_mapper[metric]



"""Metrics and visualizations for evaluating chatbot's AI utilization.

This project intents to provide techniques for evaluating the quality of a chatbot's AI knowledge base
and visualize insights from chatbot message bases.
"""

__author__ = 'squad ROps'
__email__ = 'anaytics.dar@take.net'
__version__ = '0.2.0'
__license__ = 'MIT License'
__credits__ = ['Cecilia Assis', 'Gabriel Salgado', 'Juliana Guamá', 'Milo Utsch', 'Rogers Damas']
__all__ = ['AiEvaluation']

from .domain import AiEvaluation



import matplotlib
import matplotlib.patheffects as path_effects
import matplotlib.pyplot as plt

AX = matplotlib.axes.Axes


def plot_text(text: str, title: str, title_size: int = 30) -> AX:
    """Plots a text.

    Parameters
    ----------
    text : str
        Text to be plotted.
    title : str
        Title for the output graph.
    title_size : int, optional
        Title's size (defaults is 30 pixels).

    Returns
    -------
    ax : matplotlib Axes
        Axes object with the text.
    """
    fig, ax = plt.subplots()

    text = fig.text(0.5, 0.5, text, ha='center', va='center', size=title_size * 0.9)
    text.set_path_effects([path_effects.Normal()])

    ax.set_title(title, fontsize=title_size)
    ax.set_axis_off()

    return ax



__author__ = 'Milo Utsch'
__version__ = '0.1.0'

import pytest

import pandas as pd

from take_ai_evaluation.logic import validate_ai_base


def test_validation_success():
    mock_df = pd.DataFrame([{'Sentence': 'cadastra ai', 'Intent': 'cadastrar', 'Predicted': 'cadastrar'},
                            {'Sentence': 'compra pra mim', 'Intent': 'compra', 'Predicted': 'cadastrar'},
                            ])
    mock_columns = ['Intent', 'Predicted']

    validate_ai_base(mock_df, mock_columns)
        

def test_validation_missing_col():
    mock_df = pd.DataFrame([{'Sentence': 'cadastra ai', 'entity': 'cadastrar', 'Predicted': 'cadastrar'},
                            {'Sentence': 'compra pra mim', 'entity': 'compra', 'Predicted': 'cadastrar'},
                            ])
    mock_columns = ['Intent', 'Predicted']

    with pytest.raises(KeyError, match=r'.*Dataframe does not contain required column.*'):
        validate_ai_base(mock_df, mock_columns)
        
        
def test_validation_empty_df():
    mock_df = pd.DataFrame()
    
    mock_columns = ['Intent', 'Predicted']

    with pytest.raises(Exception, match=r'.*Dataframe cannot be empty.*'):
        validate_ai_base(mock_df, mock_columns)



import os
import typing as tp

import pandas as pd

DF = pd.DataFrame
StrOrDataframe = tp.Union[str, DF]


def read_dataframe(dataframe: StrOrDataframe, sep: str = '|', encoding: str = 'utf-8') -> DF:
    """Provides a dataframe.

    If `dataframe` is a string it will read the file, otherwise it returns the passed object.

    Parameters
    ----------
    dataframe : str | pandas.DataFrame
        Dataframe object, can be a file to be read.
    sep : str, optional
        CSV separator (default is "|").
    encoding : str, optional
        File encoding (default is utf-8).

    Returns
    -------
    dataframe : pandas.DataFrame
        Read dataframe.

    Raises
    ------
    TypeError
        If `dataframe` is not the expected type.
    FileNotFoundError
        If `dataframe` not found.
    """
    if isinstance(dataframe, str):
        return read_from_file(file_name=dataframe, sep=sep, encoding=encoding)
    elif isinstance(dataframe, DF):
        return dataframe
    else:
        raise TypeError('Parameter `dataframe` must be `str` or `pandas.DataFrame`.')


def read_from_file(file_name: str, sep: str = '|', encoding: str = 'utf-8') -> DF:
    """Reads dataframe from file.

    Parameters
    ----------
    file_name : str
        Name of the file to be read.
    sep : str, optional
        CSV separator (default is "|").
    encoding : str, optional
        File encoding (default is utf-8).

    Returns
    -------
    dataframe : pandas.DataFrame
        Read dataframe.

    Raises
    ------
    FileNotFoundError
        If `file_name` not found.
    """
    if os.path.isfile(file_name):
        return pd.read_csv(filepath_or_buffer=file_name, sep=sep, encoding=encoding)
    else:
        raise FileNotFoundError(f"File at '{file_name}' not found.")



import typing as tp

import pandas as pd
import numpy as np

from .intent import Intent

DF = pd.DataFrame
DictOrDataframe = tp.Union[tp.Dict, DF]


class KnowledgeBase:
    """Wraps methods for calculating insights regarding the quality of the Knowledge Base.

    Also grants access to intent level insights.

    Attributes
    ----------
    analysed_base : str
        Analysed base file dataframe.
    labels: list of str
        List of unique intent labels in the knowledge base.
    y_true: list of str
        List of true intent labels in the knowledge base.
    y_pred: list of str
        List of predicted intent labels in the knowledge base.
    intents: dict of Intent
        Dictionary of all intents in the knowledge base.
    all_vs_all_confusion_matrix: pandas.DataFrame
        All vs ALL confusion matrix of the knowledge base.

    Methods
    -------
    get_intent(intent)
        Provides a intent.
    get_best_intent(metric)
        Provides best intent based on a metric.
    get_worst_intent(metric)
        Provides worst intent based on a metric.
    get_classification_report(as_dataframe)
        Provides metrics classification report.
    """

    def __init__(self, analysed_base: DF, intent_col: str, predict_col: str) -> None:
        """
        Parameters
        ----------
        analysed_base: pandas.DataFrame
            Analysed base file dataframe.
        intent_col : str
            Column containing the original intents in your knowledge base.
        predict_col : str
            Column containing the predicted intents in your knowledge base.
        """
        self.init_attributes(analysed_base=analysed_base, intent_col=intent_col, predict_col=predict_col)

    @property
    def analysed_base(self) -> DF:
        """Provides `analysed_base`.

        Returns
        -------
        analysed_base : pandas.DataFrame
            Dataframe object containing the knowledge base.
        """
        return self._analysed_base

    @analysed_base.setter
    def analysed_base(self, analysed_base: DF) -> None:
        """Raises an error when trying to set `analysed_base`.

        Please, do NOT use this method. Use `init_attributes` instead.

        Parameters
        ----------
        analysed_base : pandas.DataFrame
            Dataframe object, can be a file to be read.

        Raises
        ------
        AttributeError
            If `analysed_base` is trying to be set.
        """
        raise AttributeError('Please use `init_attributes` to change `analysed_base` dataset.')

    def init_attributes(self, analysed_base: DF, intent_col: str, predict_col: str) -> None:
        """Initiates all attributes for a given `analysed_base` dataset.

        Parameters
        ----------
        analysed_base: pandas.DataFrame
            Analysed base file dataframe.
        intent_col : str
            Column containing the original intents in your knowledge base.
        predict_col : str
            Column containing the predicted intents in your knowledge base.
        """
        self._analysed_base = analysed_base

        self.labels = self._analysed_base[intent_col].unique()

        self.y_true = self._analysed_base[intent_col]
        self.y_pred = self._analysed_base[predict_col]

        self.intents = {}
        self.classification_report = {}

        self.__calculate_all_vs_all_confusion_matrix()

    def __create_empty_all_vs_all_matrix(self) -> DF:
        """Initializes a ALL vs ALL (multilabel) confusion matrix filled with zeros.

        Returns
        -------
        matrix : pandas.DataFrame:
            Zero filled 2d dataframe with the size of the label count.
        """
        label_count = self.labels.size

        return pd.DataFrame(data=np.zeros((label_count, label_count)),
                            index=self.labels,
                            columns=self.labels)

    def __calculate_all_vs_all_confusion_matrix(self) -> None:
        """Calculates the confusion matrix for this knowledge base.
        """
        confusion_matrix = self.__create_empty_all_vs_all_matrix()

        for i in range(len(self.y_true)):
            confusion_matrix[self.y_true[i]][self.y_pred[i]] += 1.

        self.all_vs_all_confusion_matrix = confusion_matrix

    def get_classification_report(self) -> DictOrDataframe:
        """Provides classification metrics report.

        Returns
        -------
        classification_report : dict of dict or str.
            Classification metrics report.

        Raises
        ------
        ZeroDivisionError
            If during intent's metrics calculation the denominator evaluates to zero.
        """
        if not self.classification_report:
            self.classification_report = self.__calculate_classification_report()

        return self.classification_report

    def __calculate_classification_report(self) -> tp.Dict:
        """Calculates metrics classification report.

        Returns
        -------
        classification_report : dict of dict or str.
            Metrics classification report.

        Raises
        ------
        ZeroDivisionError
            If during intent's metrics calculation the denominator evaluates to zero.
        """
        classification_report = {label: {} for label in self.labels}

        total = self.all_vs_all_confusion_matrix.sum().sum()
        total_tp = 0.

        for label in self.labels:
            intent = self.get_intent(intent=label)

            total_tp += intent.tp
            classification_report[label]['precision'] = intent.precision
            classification_report[label]['recall'] = intent.recall
            classification_report[label]['f1-score'] = intent.f1

        classification_report['accuracy'] = total_tp / total

        return classification_report

    def get_intent(self, intent: str) -> Intent:
        """Provides intent.

        Parameters
        ----------
        intent : str
            Label of the intent to be retrieved.

        Returns
        -------
        intent : Intent
            Requested intent.

        Raises
        ------
        ZeroDivisionError
            If during intent's metrics calculation the denominator evaluates to zero.
        """
        if intent not in self.intents:
            self.intents[intent] = Intent(label=intent,
                                          all_vs_all_confusion_matrix=self.all_vs_all_confusion_matrix)

        return self.intents[intent]

    def get_best_intent(self, metric: str = 'accuracy') -> tp.Tuple[str, float]:
        """Provides best intent based on a metric.

        Parameters
        ----------
        metric : str
            Metric for which the value will be retrieved.
            Can be either of "accuracy", "precision", "recall" or "f1".

        Returns
        -------
        intent : tuple of str and float
            Tuple containing intent name and metric value.

        Raises
        ------
        KeyError
            If `metric` is not any of the expected ones.
        ZeroDivisionError
            If during intent's metrics calculation the denominator evaluates to zero.
        """
        return self.__sort_metrics(metric=metric)[-1]

    def get_worst_intent(self, metric: str = 'accuracy') -> tp.Tuple[str, float]:
        """Provides worst intent based on a metric.

        Parameters
        ----------
        metric : str
            Metric for which the value will be retrieved.
            Can be either of "accuracy", "precision", "recall" or "f1".

        Returns
        -------
        intent : tuple of str and float
            Tuple containing intent name and metric value.

        Raises
        ------
        KeyError
            If `metric` is not any of the expected ones.
        ZeroDivisionError
            If during intent's metrics calculation the denominator evaluates to zero.
        """
        return self.__sort_metrics(metric)[0]

    def __sort_metrics(self, metric: str = 'accuracy') -> tp.List[tp.Tuple]:
        """Sorts in ascending order a list of intents and its values based on a metric.

        Parameters
        ----------
        metric : str
            Metric for which the value will be retrieved.
            Can be either of "accuracy", "precision", "recall" or "f1".

        Returns
        -------
        metrics : list of tuple.
            Sorted list

        Raises
        ------
        KeyError
            If `metric` is not any of the expected ones.
        ZeroDivisionError
            If during intent's metrics calculation the denominator evaluates to zero.
        """
        metrics = {}

        for label in self.labels:
            intent = self.get_intent(intent=label)
            metrics[intent.label] = intent.get_metric(metric=metric)

        sorted_metrics = sorted(metrics.items(), key=lambda item: item[1])

        return sorted_metrics



__author__ = 'Milo Utsch and Cecília Assis'
__version__ = '0.1.0'
__all__ = ['validate_dataframe',
           'validate_parameters',
           'validate_parameter',
           'KnowledgeBase',
           'normalize_confusion_matrix',
           'Intent']

from .validation import validate_dataframe, validate_parameters, validate_parameter
from .knowledge_base import KnowledgeBase
from .utils import normalize_confusion_matrix
from .intent import Intent



from setuptools import setup, find_packages

from take_ai_evaluation import __name__ as name
from take_ai_evaluation import __author__ as author
from take_ai_evaluation import __doc__ as doc
from take_ai_evaluation import __email__ as author_email
from take_ai_evaluation import __version__ as version
from take_ai_evaluation import __license__ as license
from take_ai_evaluation import __credits__ as credits

maintainer = "daresearch"
maintainer_email = "anaytics.dar@take.net"
keywords = ['ai-knowledge', 'chatbot', 'classification', 'evaluation']
description = doc.splitlines()[0].strip()
long_description = open('README.md').read()
long_description_content_type = 'text/markdown'
install_requires = [
    element
    for element in [
        line.split('#', 1)[0].strip()
        for line in open('requirements.txt', 'r', encoding='utf-8')
    ]
    if element and not element.startswith('--')
]
classifiers = [
    'Programming Language :: Python :: 3',
    'Operating System :: OS Independent'
]

setup(
    name=name,
    author=author,
    author_email=author_email,
    version=version,
    license=license,
    credits=credits,
    maintainer=maintainer,
    maintainer_email=maintainer_email,
    keywords=keywords,
    description=description,
    long_description=long_description,
    long_description_content_type=long_description_content_type,
    install_requires=install_requires,
    classifiers=classifiers,
    packages=find_packages(exclude=['tests'])
)



import pandas as pd
import seaborn as sns
import matplotlib

from take_ai_evaluation.logic import normalize_confusion_matrix

DF = pd.DataFrame
AX = matplotlib.axes.Axes


def plot_confusion_matrix(confusion_matrix: DF,
                          title: str,
                          xlabel: str = '',
                          ylabel: str = '',
                          normalize: bool = False) -> AX:
    """Provides the confusion matrix for a knowledge base.

    Parameters
    ----------
    confusion_matrix : pandas.DataFrame
        Confusion matrix to be plotted.
    title : str
        Title for the output graph.
    xlabel : str, optional
        X axis label for the output graph (default is '').
    ylabel : str, optional
        Y axis label for the output graph (default is '').
    normalize : bool, optional
        If True, return output normalized using L1 normalization (default is false).

    Returns
    -------
    ax : matplotlib Axes
        Axes object with the heatmap.
    """
    if normalize:
        confusion_matrix = normalize_confusion_matrix(confusion_matrix=confusion_matrix)

    return _plot_confusion_matrix(confusion_matrix=confusion_matrix, title=title, xlabel=xlabel, ylabel=ylabel)


def _plot_confusion_matrix(confusion_matrix: pd.DataFrame, title: str, xlabel: str, ylabel: str) -> AX:
    """Provides the confusion matrix for a knowledge base.

    Parameters
    ----------
    confusion_matrix : pandas.DataFrame
        Confusion matrix to be plotted.
    title : str
        Title for the output graph.
    xlabel : str, optional
        X axis label for the output graph (default is '').
    ylabel : str, optional
        Y axis label for the output graph (default is '').

    Returns
    -------
    ax : matplotlib Axes
        Axes object with the heatmap.
    """
    ax = sns.heatmap(data=confusion_matrix, annot=True, linewidths=.5)

    ax.set_title(label=title)
    ax.set_xlabel(xlabel=xlabel)
    ax.set_ylabel(ylabel=ylabel)

    return ax



import typing as tp

import pandas as pd
import matplotlib

from take_ai_evaluation.logic import KnowledgeBase, validate_dataframe, validate_parameters, validate_parameter
from take_ai_evaluation.persistence import read_dataframe
from take_ai_evaluation.presentation import plot_confusion_matrix, plot_text

DF = pd.DataFrame
AX = matplotlib.axes.Axes

StrOrDataframe = tp.Union[str, DF]
TupleOrAxes = tp.Union[tp.Tuple, AX]
DictOrDataframe = tp.Union[tp.Dict, DF]


class AiEvaluation:
    """Wraps methods for calling and using the package's visualizations.

    Attributes
    ----------
    intent_col : str
        Column containing the original intents in your knowledge base.
    predict_col: str
        Column containing the original intents in your knowledge base.
    analysed_base: pandas.DataFrame
        Dataframe object containing the knowledge base.
    knowledge_base: KnowledgeBase
        Object that wraps methods for calculating insights regarding the quality of the Knowledge Base.

    Methods
    -------
    init_attributes(analysed_base, sentence_col, intent_col, predict_col, analysed_base_sep)
        Initiates all attributes for a given `analysed_base` dataset.
    get_all_vs_all_confusion_matrix(title, normalize=False)
        Provides the ALL vs ALL confusion matrix.
    get_one_vs_all_confusion_matrix(title, normalize=False)
        Provides the ONE vs ALL confusion matrix.
    get_best_intent(metric='accuracy', as_graph=False)
        Provides best intent based on a metric.
    get_worst_intent(metric='accuracy', as_graph=False)
        Provides worst intent based on a metric.
    get_classification_report(as_dataframe=False)
        Provides classification metrics report.

    Raises
    ------
    TypeError
        If the during `analysed_base` validation any parameter's type or itself does not match the desired type.
    ValueError
        If the during `analysed_base` validation any parameter or itself is empty.
    FileNotFoundError
        If during knowledge base reading `analysed_base` file is not found.
    AttributeError
        If any of the expected columns is not present inside the knowledge base dataframe.
    """

    def __init__(self,
                 analysed_base: StrOrDataframe,
                 sentence_col: str,
                 intent_col: str,
                 predict_col: str,
                 analysed_base_sep: str = '|') -> None:
        """
        Parameters
        ----------
        analysed_base : str | pandas.DataFrame
            Dataframe object, can be a file to be read.
        sentence_col : str
            Column containing the original sentence in your knowledge base.
        intent_col : str
            Column containing the original intents in your knowledge base.
        predict_col : str
            Column containing the predicted intents in your knowledge base.
        analysed_base_sep : str, optional
            CSV separator (default is "|").
        """
        self.init_attributes(analysed_base=analysed_base,
                             sentence_col=sentence_col,
                             intent_col=intent_col,
                             predict_col=predict_col,
                             analysed_base_sep=analysed_base_sep)

    def __validate_params(self,
                          analysed_base: StrOrDataframe,
                          sentence_col: str,
                          intent_col: str,
                          predict_col: str) -> None:
        """Validates parameters' types and values.

        Parameters
        ----------
        analysed_base : str | pandas.DataFrame
            Dataframe object, can be a file to be read.
        sentence_col : str
            Column containing the original sentence in your knowledge base.
        intent_col : str
            Column containing the original intents in your knowledge base.
        predict_col : str
            Column containing the predicted intents in your knowledge base.

        Raises
        ------
        TypeError
            If the parameter's type does not match the desired type.
        ValueError
            If the parameter is empty.
        """
        validate_parameter(param=analysed_base, param_name='analysed_base', target_type=(str, DF))
        validate_parameter(param=sentence_col, param_name='sentence_col', target_type=str)
        validate_parameter(param=intent_col, param_name='intent_col', target_type=str)
        validate_parameter(param=predict_col, param_name='predict_col', target_type=str)

    @property
    def analysed_base(self) -> DF:
        """Provides `analysed_base`.

        Returns
        -------
        analysed_base : pandas.DataFrame
            Dataframe object containing the knowledge base.
        """
        return self._analysed_base

    @analysed_base.setter
    def analysed_base(self, analysed_base: StrOrDataframe) -> None:
        """Raises an error when trying to set `analysed_base`.

        Please, do NOT use this method. Use `init_attributes` instead.

        Parameters
        ----------
        analysed_base : pandas.DataFrame
            Dataframe object, can be a file to be read.

        Raises
        ------
        AttributeError
            If `analysed_base` is trying to be set.
        """
        raise AttributeError('Please use `init_attributes` to change `analysed_base` dataset.')

    def init_attributes(self,
                        analysed_base: StrOrDataframe,
                        sentence_col: str,
                        intent_col: str,
                        predict_col: str,
                        analysed_base_sep: str) -> None:
        """Initiates all attributes for a given `analysed_base` dataset.

        Parameters
        ----------
        analysed_base : str | pandas.DataFrame
            Dataframe object, can be a file to be read.
        sentence_col : str
            Column containing the original sentence in your knowledge base.
        intent_col : str
            Column containing the original intents in your knowledge base.
        predict_col : str
            Column containing the predicted intents in your knowledge base.
        analysed_base_sep : str
            CSV separator.

        Raises
        ------
        TypeError
            If the during `analysed_base` validation any parameter's type or itself does not match the desired type.
        ValueError
            If the during `analysed_base` validation any parameter or itself is empty.
        FileNotFoundError
            If during knowledge base reading `analysed_base` file is not found.
        AttributeError
            If any of the expected columns is not present inside the knowledge base dataframe.
        """
        self.__validate_params(analysed_base=analysed_base,
                               sentence_col=sentence_col,
                               intent_col=intent_col,
                               predict_col=predict_col)

        self.intent_col = intent_col
        self.predict_col = predict_col

        self.__init_analysed_base(analysed_base=analysed_base,
                                  analysed_base_sep=analysed_base_sep,
                                  sentence_col=sentence_col)

        self.__init_knowledge_base()

    def __init_analysed_base(self, analysed_base: StrOrDataframe, analysed_base_sep: str, sentence_col: str) -> None:
        """Reads knowledge base based on `analysed_base`.

        Parameters
        ----------
        analysed_base : str | pandas.DataFrame
            Dataframe object, can be a file to be read.
        analysed_base_sep : str
            CSV separator.
        sentence_col : str
            Column containing the original sentence in your knowledge base.

        Raises
        ------
        TypeError
            If `analysed_base` is not the expected type.
        FileNotFoundError
            If `analysed_base` file is not found.
        ValueError
            If after read `analysed_base` is empty.
        AttributeError
            If during `analysed_base` validation any of the expected columns is not present in it.
        """
        analysed_base = read_dataframe(dataframe=analysed_base, sep=analysed_base_sep)

        columns = [sentence_col, self.intent_col, self.predict_col]
        validate_dataframe(df=analysed_base, columns=columns)

        self._analysed_base = analysed_base

    def __init_knowledge_base(self) -> None:
        """Initializes the knowledge base object after reading it's files.
        """
        self.knowledge_base = KnowledgeBase(analysed_base=self._analysed_base,
                                            intent_col=self.intent_col,
                                            predict_col=self.predict_col)

    def get_all_vs_all_confusion_matrix(self, title: str, normalize: bool = False) -> AX:
        """Provides the ALL vs ALL confusion matrix for a knowledge base.

        Parameters
        ----------
        title : str
            Wanted title for the output graph.
        normalize : bool, optional
            If True, return output normalized using L1 normalization (default is false).

        Returns
        -------
        matrix : matplotlib Axes
            Axes object with the confusion matrix.
        """
        return plot_confusion_matrix(confusion_matrix=self.knowledge_base.all_vs_all_confusion_matrix,
                                     title=title,
                                     xlabel='Predicted label',
                                     ylabel='True label',
                                     normalize=normalize)

    def get_one_vs_all_confusion_matrix(self, intent: str, title: str, normalize: bool = False) -> AX:
        """Provides the ONE vs ALL confusion matrix for a knowledge base and a given intent.

        Parameters
        ----------
        intent : str
            Label of the intent to be retrieved.
        title : str
            Wanted title for the output graph.
        normalize : bool, optional
            If True, return output normalized using L1 normalization (default is false).

        Returns
        -------
        matrix : matplotlib Axes
            Axes object with the confusion matrix.

        Raises
        ------
        ZeroDivisionError
            If during intent's metrics calculation the denominator evaluates to zero.
        """
        intent = self.knowledge_base.get_intent(intent=intent)

        return plot_confusion_matrix(confusion_matrix=intent.one_vs_all_confusion_matrix,
                                     title=title,
                                     xlabel='Predicted label',
                                     ylabel='True label',
                                     normalize=normalize)

    def get_worst_intent(self, metric: str = 'accuracy', as_graph: bool = False) -> TupleOrAxes:
        """Provides worst intent based on a metric.

        Parameters
        ----------
        metric : str, optional
            Metric to be evaluated against (default is accuracy).
            Can be either of "accuracy", "precision", "recall" or "f1".
        as_graph : bool, optional
            If True, shows the result as a graph, i.e. image (default is false).

        Returns
        -------
        intent : tuple or matplotlib Axes
            Worst intent's name and value.

        Raises
        ------
        KeyError
            If `metric` is any of the expected ones.
        """
        intent, value = self.knowledge_base.get_worst_intent(metric=metric)

        if as_graph:
            text = f'Intent: {intent}\n{metric.capitalize()}: {value:.2f}'
            return plot_text(text=text, title='Worst intent')

        return intent, value

    def get_best_intent(self, metric: str = 'accuracy', as_graph: bool = False) -> TupleOrAxes:
        """Provides best intent based on a metric.

        Parameters
        ----------
        metric : str, optional
            Metric to be evaluated against (default is accuracy).
            Can be either of "accuracy", "precision", "recall" or "f1".
        as_graph : bool, optional
            If True, shows the result as a graph, i.e. image (default is false).

        Returns
        -------
        intent : tuple or matplotlib Axes
            Best intent's name and value.

        Raises
        ------
        KeyError
            If `metric` is any of the expected ones.
        """
        intent, value = self.knowledge_base.get_best_intent(metric=metric)

        if as_graph:
            text = f'Intent: {intent}\n{metric.capitalize()}: {value:.2f}'
            return plot_text(text=text, title='Best intent')

        return intent, value

    def get_classification_report(self, as_dataframe: bool = False) -> DictOrDataframe:
        """Provides classification metrics report.

        Parameters
        ----------
        as_dataframe : bool, optional
            If True, return output as pandas dataframe, otherwise return as dict (default is false).

        Returns
        -------
        classification_report : dict of dict or str | pandas DataFrame.
            Classification metrics report.

        Raises
        ------
        ZeroDivisionError
            If during intent's metrics calculation the denominator evaluates to zero.
        """
        classification_report = self.knowledge_base.get_classification_report()

        if as_dataframe:
            dataframe = pd.DataFrame(data=classification_report).transpose()
            dataframe.loc['accuracy'].iloc[:-1] = 0.

            return dataframe

        return classification_report
